<?php

require_once(__DIR__."/sphinxapi.php");

class Search {
	protected $cl;
	protected $starttime;


	public function __construct() {
		$this->cl = new SphinxClient();
		$this->cl->SetServer(DocConfig::get('search_host'), DocConfig::get('search_port'));
		$this->cl->SetMatchMode(SPH_MATCH_EXTENDED);
		$this->cl->SetConnectTimeout(3);
		$this->cl->SetArrayResult(true);
		$this->starttime = microtime();
		$this->prepare_templates("templates");
	}

	protected function prepare_templates() {
		$content = DocCache::getTemplate("search.html");

		$tpls = array();
		preg_match_all("/====madoka_tpl_begin====:(.*)\n(.*)====madoka_tpl_end====/sU", $content, $tpls);
		$this->tpls = array();
		for ($i = 0; $i < count($tpls[1]); $i++) {
			$name = trim($tpls[1][$i]);
			$tpl = trim($tpls[2][$i]);
			$this->tpls[$name] = $tpl;
		}
	}
	protected function template($name){
		return $this->tpls[$name];
	}

	public function replace_vars($content, $vars, $value = false) {
		if ($value !== false)
			$this->vars = array($vars => $value);
		else
			$this->vars = $vars;
		$content = preg_replace_callback('/\$\$([a-zA-Z0-9\\.\\_\\-\\!]+)?/', array($this, "replace_vars_callback"), $content);
		return $content;
	}

	public function replace_vars_callback($matches) {
		$varname = $matches[1];
		if (isset($this->vars[$varname]))
			return $this->vars[$varname];
		
		return "";
	}

	protected function query($query, $index, $limit_start = 0, $limit_count = 10) {
		if ($query === false) return false;
		if ($limit_start < 0) $limit_start = 0;
		if ($limit_count < 0) $limit_count *= -1;
		$this->cl->setLimits($limit_start, $limit_count);
		$query = $this->prepare_word($query);
		$res = $this->cl->Query($query, $index);
		return $res;
	}

	protected function prepare_word($query) {
		$query = trim($query);
		$query = preg_split("/[^\w\d]/", $query);
		for ($i = 0; $i < count($query); $i++)
			$query[$i] = "({$query[$i]}|{$query[$i]}*)";
		$query = implode('&', $query);
		return $query;
	}

	protected function links($res) {
		$links = array();
		if ($res === false || !(isset($res['matches']))) return $links;
		$pages = $res['matches'];
		for ($i = 0; $i < count($pages); $i++) {
			$links[] = $pages[$i]['attrs'];
		}
		return $links;
	}

	protected function total($res) {
		return (isset($res['total_found'])) ? $res['total_found'] : 0;
	}



	protected function paging($page, $total, $per_page) {
		$number = ceil($total/$per_page);
		if ($number > 10) $number = 10;
		if ($total <= $per_page) return "";

		$tpl_normal = $this->template('paging_normal');
		$tpl_active = $this->template('paging_active');
		$link = 'search.php?query='.urlencode($_GET["query"]).'&page=';

		$items = array();
		for ($i = 0; $i < $number; $i++) {
			$tpl = ($i == $page) ? $tpl_active : $tpl_normal;
			$item = $this->replace_vars($tpl, array(
				'link' => $link.($i),
				'number' => $i + 1
			));
			$items[] = $item;
		}


		$tpl = $this->template('paging');
		$paging = $this->replace_vars($tpl, array(
			'links' => implode("", $items)
		));
		return $paging;
	}

	protected function paging_btn($type, $link, $condition) {
		if ($condition) {
			$tpl = $this->template('paging_btn');
			$tpl = $this->replace_vars($tpl, array(
				'link' => "<a href=\"$link\"></a>",
				'type' => $type
			));
		} else {
			$tpl = $this->template('paging_btn');
			$tpl = $this->replace_vars($tpl, array(
				'link' => '<a></a>',
				'type' => $type.' inactive'
			));
		}
		return $tpl;
	}

	public function get_components($query, $count = 5){
		$result = array();
		$stops = DocConfig::get("search_components");
		if ($stops){
			$squery = strtolower($query);
			$stops = explode(",", $stops);
			for ($i=0; $i < sizeof($stops); $i++) { 
				if (sizeof($result) >= $count) break;
				if (strpos($stops[$i], $query) !== false){
					$name = str_replace("ui.", "", $stops[$i]);
					$result[] = array(
						"short"	=>	"API reference for ".$stops[$i]." component",
						"name"  =>	"api__refs__".$stops[$i].".html",
						"label"	=>	$stops[$i],
						"icon"  =>  $name,
						"type"	=>	"none"
					);
				}
			}
		}

		return $result;
	}

	public function suggest($query){
		$total = 0;

		//API
		$this->cl->setGroupBy("title", SPH_GROUPBY_ATTR);
		$data = $this->query($query, DocConfig::get("name")."_api", 0 , 5);
		$this->cl->resetGroupBy();

		$matches = $this->links($data);

		$data2 = $this->query($query, DocConfig::get("name")."_pages", 0 , 5);
		$matches = array_merge($matches, $this->links($data2));
		
		if (!count($matches))
			return $matches;

		$result = $this->get_components($query);
		
		for ($i=0; $i<count($matches); $i++){
			if (sizeof($result) > 4) break;

			$omatch = $matches[$i];
			$short = isset($matches[$i]["short"]) ? $omatch["short"] : "";
			if (DocConfig::get("api_signature") && isset($matches[$i]["signature"]))
				$label = $omatch["signature"];
			else
				$label = $omatch["title"];

			$result[] = array(
				"name"  => $omatch["link"],
				"label" => $label,
				"short" => $short,
				"type"  => @$omatch["type"]
			);
		}

		$total = $this->total($data) + $this->total($data2);
		if ($total > 5)
			$result[] = array(
				"name" => "search.php?query=".$query,
				"label" => "and $total more results",
				"short" => "", "type" => "total"
			);

		return $result;
	}

	public function render() {
		$content = DocCache::getTemplate('layout.html');
		$query = $_GET["query"];

		if (isset($_GET['viewall'])) {
			if ($_GET['viewall'] === "api") {
				$articles = $samples = '';
				$api = $this->render_api($query, true);
			} else if ($_GET['viewall'] === "samples") {
				$articles = $api = '';
				$samples = $this->render_samples($query, true);
			}
		} else {
			$components = $this->render_components($query);
			$articles = $this->render_pages($query);
			if (!isset($_GET["page"]) || $_GET["page"] == 0){
				if (DocConfig::get("sample_results"))
					$samples = $this->render_samples($query);
				else 
					$samples = "";

				if (DocConfig::get("api_results"))
					$api = $this->render_api($query);
				else
					$api = "";
			} else {
				$api = $samples = "";
			}
		}

		$data = $components.$api.$articles.$samples;

		if (strlen($data) === 0)
			$data = $this->template('nothing');

		
		$pageData = DocConfig::$data;
		$pageData["data"] = $data;
		$pageData["rendertime"] = $this->rendertime($query);
		$pageData["pageclass"] = "search";
		$pageData["!header.html"] = DocCache::getTemplate('header.html');
		$pageData["!footer.html"] = DocCache::getTemplate('footer.html');

		$content = $this->replace_vars($content, $pageData);

		return $content;
	}

	
	protected function rendertime($query) {
		$rendertime = microtime() - $this->starttime;
		if ($rendertime < 0) $rendertime = $this->faketime();
		$render = $this->template('rendertime');
		return $this->replace_vars($render, array(
			'rendertime' => round($rendertime, 3),
			'search_word' => $query
		));
	}

	protected function faketime() {
		$rendertime = '0';
		$time = rand(1, 20)/102;
		$time = sprintf("%.5f", $time);
		return $time;
	}

	protected function render_components($query){
		$results = $this->get_components($query, 5);
		for ($i = 0; $i < count($results); $i++) {
			$name = $results[$i]["label"];
			$link = $results[$i]["name"];
			$icon = $results[$i]["icon"];
			$results[$i] = "<div class='search_ui'><a href='".$link."'><img src='./media/api/toc/".$icon.".png'><br>".$name."</a></div>";
		}

		return implode("", $results);
	}

	protected function render_pages($query) {
		$page = isset($_GET['page']) ? (int) $_GET['page'] : 0;
		$page = abs($page);

		$tpl = $this->template('articles');
		$item_tpl = $this->template('article_item');

		$this->cl->SetSortMode(SPH_SORT_ATTR_ASC, "api");
		$res = $this->query($query, DocConfig::get("name")."_pages", $page*DocConfig::get("search_results"), DocConfig::get("search_results"));
		$this->cl->SetSortMode(SPH_SORT_RELEVANCE);

		$pages = $this->links($res);
		$total = $this->total($res);
		
		$paging = $this->paging($page, $total, DocConfig::get("search_results"));
		if (count($pages) === 0) return "";

		for ($i = 0; $i < count($pages); $i++) {
			$content = file_get_contents(DocConfig::get("search_root").$pages[$i]['link']);
			$pagecontent = $this->page_content($content);

			$pages[$i] = $this->replace_vars($item_tpl, array(
				'link' => $pages[$i]['link'],
				'category' => $this->page_crumbs($content),
				'title' => $pages[$i]['title'],
				'content' => $this->textwrap($pagecontent, $query)
			));
		}

		$result = $this->render_details($tpl, array(
			'dataset' => implode("", $pages),
			'paging' => $paging
		), count($pages), $page, DocConfig::get("search_results"), $total, 'articles');

		return $result;
	}

	protected function listComponents($usedby){
		$parts = explode(",", $usedby);
		$count = count($parts)-1;
		$docname = DocConfig::get("name");

		for ($i=$count; $i >= 0; $i--) { 
			$temp = explode("__", $parts[$i]);
			$temp = explode("_", array_pop($temp));
			$name = str_replace($docname, "", $temp[0]);
			if ($name == "")
				$name = $docname;

			$parts[$i] = "<a href='".$parts[$i]."'>".$name."</a>";
		}

		$text = implode(" ", $parts);
		if ($count > 6)
			return "<a href='#' onclick='this.style.display=\"none\"; this.nextSibling.style.display = \"block\"; return false;'>More than 6 components, click to see all</a><div style='display:none'>".$text."</div>";

		return $text;
	}

	protected function render_api($query, $all = false) {
		$tpl = $this->template('api');
		$item_tpl = $this->template('api_item');
		$limit = $all ? 1000 : DocConfig::get("api_results");

		$this->cl->setGroupBy("title", SPH_GROUPBY_ATTR);
		$res = $this->query($query, DocConfig::get("name")."_api", 0, $limit);
		$this->cl->resetGroupBy();
		$pages = $this->links($res);
		$total = $this->total($res);
		if (count($pages) === 0) return "";

		for ($i = 0; $i < count($pages); $i++) {
			$usedby = "";
			if (DocConfig::get("api_usedby"))
				$usedby = $this->listComponents( $pages[$i]['usedby'] );

			$vars = array(
				'short' => $pages[$i]['short'],
				'link' => $pages[$i]['link'],
				'title' => $pages[$i]['title'],
				'icon' => $pages[$i]['type'],
				'usedby' => $usedby
			);
			$pages[$i] = $this->replace_vars($item_tpl, $vars);
		}

		$per_page = $all ? $total : DocConfig::get("api_results");

		$result = $this->render_details($tpl, array(
			"dataset" => implode("", $pages)
		), count($pages), 0, $per_page, $total, 'api', $all);
		return $result;
	}

	protected function render_samples($query, $all = false) { 
		$tpl = $this->template('samples');
		$item_tpl = $this->template('samples_item');
		$limit = $all ? 1000 : DocConfig::get("sample_results");
		$res = $this->query($query, DocConfig::get("name")."_samples", 0, $limit);
		$pages = $this->links($res);
		$total = $this->total($res);
		
		if (count($pages) === 0) return "";

		$url = DocConfig::get("samples_http");

		for ($i = 0; $i < count($pages); $i++) {
			$pages[$i] = $this->replace_vars($item_tpl, array(
				'link' => $url.$pages[$i]['link'],
				'title' => $pages[$i]['title']
			));
		}

		$per_page = $all ? $total : DocConfig::get("sample_results");

		$result = $this->render_details($tpl, array(
			"dataset" => implode("", $pages)
		), count($pages), 0, $per_page, $total, 'samples');

		return $result;
	}

	
	protected function render_details($content, $data, $pages, $page, $per_page, $total, $type, $no_viewall = false) {
		if ($pages < $total && !$no_viewall) {
			$link = 'search.php?query='.urlencode($_GET["query"]).'&viewall='.$type;
			$viewall = $this->replace_vars($this->template('viewall'), array( 'link' => $link ));
		} else
			$viewall = "";

		$data["viewall"] = 		$viewall;
		$data["shown"] = 		$this->render_per_page($page, $per_page, $total);
		$data["total_found"] = 	$total;
		$data["entries"] = 		($pages > 1) ? 'entries are shown' : 'entry is shown';

		$result = $this->replace_vars($content, $data);
		return $result;
	}

	protected function render_per_page($page, $per_page, $total) {
		$start = $page*$per_page + 1;
		$end = min(($page + 1)*$per_page, $total);
		if ($start == $end) return $start;
		return $start." - ".$end;
	}

	protected function page_crumbs($content){
		$br_start = strpos($content, '<a class="n1"');
		$br_end = strpos($content, "</div>", $br_start);

		$str =  substr($content, $br_start, $br_end-$br_start);
		if (!$str) return "";

		$parts = explode("<a", $str);

		if (strpos($str, "API Reference") !== false){
			if (sizeof($parts) > 2)
				return "<a".$parts[sizeof($parts)-2];
			else
				return "<a".$parts[sizeof($parts)-1];
		}

		return "<a".$parts[sizeof($parts)-1];
	}
	protected function page_content($content) {
		$result = $content;


		$start = strpos($content, "<!-- Content Area -->");
		$end   = strpos($content, "<!-- Content Area End -->");

		$result = substr($content, $start, $end-$start);

		$result = preg_replace("/<pre>.*<code>.*<\/code>.*<\/pre>/Us", "", $result);
		$result = preg_replace("/<script.*?<\/script>/Us", "", $result);
		$result = preg_replace("/<h1>.*<\/h1>/i", "", $result);

		return $result;
	}


	protected function textwrap($text, $key, $words_limit = 8, $keys_limit = 2) {
		$result = array();
		$text = $this->strip_tags($text);
		$text = str_replace("\n", " ", $text);
		$text = str_replace("\r", " ", $text);
		$text = str_replace("\t", " ", $text);

		$regexp = "/\w*".str_replace("/", "\\/", $key)."\w*/i";
		$keys = array();
		preg_match_all($regexp, $text, $keys);
		if (!isset($keys[0])) return $text;
		$keys = $keys[0];
		for ($i = 0; $i < count($keys); $i++)
			$keys[$i] = "<span class=\"search_select\">{$keys[$i]}</span>";

		$parts = preg_split($regexp, $text);

		for ($i = 0; $i < count($parts); $i++) {
			$regsep = "/[ ,.]+/";
			$part = $parts[$i];
			$words = preg_split($regsep, $part);

			if (count($words) <= $words_limit*2) continue;
			$sep = array();
			preg_match_all($regsep, $part, $sep);
			$sep = $sep[0];

			$len_start = 0;
			for ($j = 0; $j < count($words) && $j <= $words_limit; $j++)
				$len_start += strlen($words[$j]) + (isset($sep[$j]) ? strlen($sep[$j]) : 0);

			$len_end = 0;
			$empty = 0;
			for ($j = count($words) - 1; $j >=0 && count($words) - $j <= $words_limit + $empty; $j--) {
				if (empty($words[$j])) $empty++;
				$len_end += strlen($words[$j]) + (isset($sep[$j - 1]) ? strlen($sep[$j - 1]) : 0);
			}
			$parts[$i] = substr($part, 0, $len_start)." ... ".substr($part, $len_end*-1);
		}
		$text = "";
		for ($i = 0; $i < count($keys) && $i < $keys_limit; $i++)
			$text .= $parts[$i].$keys[$i];
		$text .= $parts[$i];
		return $text;
	}

	protected function strip_tags($text) {
		$text = preg_replace("/<[^>]+>/s", " ", $text);
		return $text;
	}

}

?>